@@ -20,6 +20,7 @@
  */
 
 #include <glib.h>
+#include <stdlib.h>
 #include <string.h>
 
 #include <locale.h>
@@ -31,15 +32,33 @@
 #endif
 
 #include <pango/pangocairo.h>
+
+#ifdef G_OS_WIN32
+#include <pango/pangowin32.h>
+#endif
+
+#ifdef HAVE_FREETYPE
+#include <pango/pangoft2.h>
+#include <pango/pangocairo-fc.h>
+#endif
+
 #include "test-common.h"
 
+
 char *
 diff_with_file (const char  *file,
                 char        *text,
                 gssize       len,
                 GError     **error)
 {
+#ifdef G_OS_WIN32
+  const char *command[] = { "diff", "-u", "-i", "--strip-trailing-cr", file, NULL, NULL };
+  const int tmp_file_idx = 5;
+#else
   const char *command[] = { "diff", "-u", "-i", file, NULL, NULL };
+  const int tmp_file_idx = 4;
+#endif
+
   char *diff, *tmpfile;
   int fd;
 
@@ -62,7 +81,7 @@ diff_with_file (const char  *file,
       goto done;
     }
   close (fd);
-  command[4] = tmpfile;
+  command[tmp_file_idx] = tmpfile;
 
   /* run diff command */
   g_spawn_sync (NULL,
@@ -81,77 +100,100 @@ done:
   return diff;
 }
 
+char *
+diff_bytes (GBytes  *b1,
+            GBytes  *b2,
+            GError **error)
+{
+  const char *command[] = { "diff", "-u", "-i", NULL, NULL, NULL };
+  char *diff, *tmpfile, *tmpfile2;
+  int fd;
+  const char *text;
+  gsize len;
+
+  /* write the text buffer to a temporary file */
+  fd = g_file_open_tmp (NULL, &tmpfile, error);
+  if (fd < 0)
+    return NULL;
+
+  text = (const char *) g_bytes_get_data (b1, &len);
+  if (write (fd, text, len) != (int) len)
+    {
+      close (fd);
+      g_set_error (error,
+                   G_FILE_ERROR, G_FILE_ERROR_FAILED,
+                   "Could not write data to temporary file '%s'", tmpfile);
+      goto done;
+    }
+  close (fd);
+
+  fd = g_file_open_tmp (NULL, &tmpfile2, error);
+  if (fd < 0)
+    return NULL;
+
+  text = (const char *) g_bytes_get_data (b2, &len);
+  if (write (fd, text, len) != (int) len)
+    {
+      close (fd);
+      g_set_error (error,
+                   G_FILE_ERROR, G_FILE_ERROR_FAILED,
+                   "Could not write data to temporary file '%s'", tmpfile2);
+      goto done;
+    }
+  close (fd);
+
+  command[3] = tmpfile;
+  command[4] = tmpfile2;
+
+  /* run diff command */
+  g_spawn_sync (NULL,
+                (char **) command,
+                NULL,
+                G_SPAWN_SEARCH_PATH,
+                NULL, NULL,
+                &diff,
+                NULL, NULL,
+                error);
+
+done:
+  unlink (tmpfile);
+  g_free (tmpfile);
+  unlink (tmpfile2);
+  g_free (tmpfile2);
+
+  return diff;
+}
+
+gboolean
+file_has_prefix (const char  *filename,
+                 const char  *str,
+                 GError     **error)
+{
+  char *contents;
+  gsize len;
+  gboolean ret;
+
+  if (!g_file_get_contents (filename, &contents, &len, error))
+    return FALSE;
+
+  ret = g_str_has_prefix (contents, str);
+
+  g_free (contents);
+
+  return ret;
+}
+
 void
 print_attribute (PangoAttribute *attr, GString *string)
 {
-  GEnumClass *class;
-  GEnumValue *value;
+  PangoAttrList *l = pango_attr_list_new ();
+  char *s;
 
-  g_string_append_printf (string, "[%d,%d]", attr->start_index, attr->end_index);
-
-  class = g_type_class_ref (pango_attr_type_get_type ());
-  value = g_enum_get_value (class, attr->klass->type);
-  g_string_append_printf (string, "%s=", value->value_nick);
-  g_type_class_unref (class);
-
-  switch (attr->klass->type)
-    {
-    case PANGO_ATTR_LANGUAGE:
-      g_string_append_printf (string, "%s", pango_language_to_string (((PangoAttrLanguage *)attr)->value));
-      break;
-    case PANGO_ATTR_FAMILY:
-    case PANGO_ATTR_FONT_FEATURES:
-      g_string_append_printf (string, "%s", ((PangoAttrString *)attr)->value);
-      break;
-    case PANGO_ATTR_STYLE:
-    case PANGO_ATTR_WEIGHT:
-    case PANGO_ATTR_VARIANT:
-    case PANGO_ATTR_STRETCH:
-    case PANGO_ATTR_SIZE:
-    case PANGO_ATTR_ABSOLUTE_SIZE:
-    case PANGO_ATTR_UNDERLINE:
-    case PANGO_ATTR_OVERLINE:
-    case PANGO_ATTR_STRIKETHROUGH:
-    case PANGO_ATTR_RISE:
-    case PANGO_ATTR_FALLBACK:
-    case PANGO_ATTR_LETTER_SPACING:
-    case PANGO_ATTR_GRAVITY:
-    case PANGO_ATTR_GRAVITY_HINT:
-    case PANGO_ATTR_FOREGROUND_ALPHA:
-    case PANGO_ATTR_BACKGROUND_ALPHA:
-    case PANGO_ATTR_ALLOW_BREAKS:
-    case PANGO_ATTR_INSERT_HYPHENS:
-    case PANGO_ATTR_SHOW:
-      g_string_append_printf (string, "%d", ((PangoAttrInt *)attr)->value);
-      break;
-    case PANGO_ATTR_FONT_DESC:
-      {
-        char *text = pango_font_description_to_string (((PangoAttrFontDesc *)attr)->desc);
-        g_string_append_printf (string, "%s", text);
-        g_free (text);
-      }
-      break;
-    case PANGO_ATTR_FOREGROUND:
-    case PANGO_ATTR_BACKGROUND:
-    case PANGO_ATTR_UNDERLINE_COLOR:
-    case PANGO_ATTR_OVERLINE_COLOR:
-    case PANGO_ATTR_STRIKETHROUGH_COLOR:
-      {
-        char *text = pango_color_to_string (&((PangoAttrColor *)attr)->color);
-        g_string_append_printf (string, "%s", text);
-        g_free (text);
-      }
-      break;
-    case PANGO_ATTR_SHAPE:
-      g_string_append_printf (string, "shape");
-      break;
-    case PANGO_ATTR_SCALE:
-      g_string_append_printf (string,"%f", ((PangoAttrFloat *)attr)->value);
-      break;
-    default:
-      g_assert_not_reached ();
-      break;
-    }
+  pango_attr_list_insert (l, pango_attribute_copy (attr));
+  s = pango_attr_list_to_string (l);
+  g_string_append (string, s);
+  g_free (s);
+  pango_attr_list_unref (l);
 }
 
 void
@@ -210,3 +252,33 @@ get_script_name (GUnicodeScript s)
   return nick;
 }
 
+/* Create a fontmap that will return our included Cantarell-VF.otf
+ * for "Cantarell"
+ *
+ * FIXME: Make this work for macOS
+ */
+PangoFontMap *
+get_font_map_with_cantarell (void)
+{
+  PangoFontMap *fontmap;
+  GError *error = NULL;
+  char *path;
+
+  fontmap = pango_cairo_font_map_new ();
+
+  /* FIXME: the coretext backend doesn't implement add_font_file */
+  if (strcmp (G_OBJECT_TYPE_NAME (fontmap), "PangoCairoCoreTextFontMap") != 0)
+    {
+      path = g_test_build_filename (G_TEST_DIST, "fonts", "Cantarell-VF.otf", NULL);
+      if (g_test_verbose ())
+        g_test_message ("adding %s to font map", path);
+      pango_font_map_add_font_file (fontmap, path, &error);
+      g_assert_no_error (error);
+      g_free (path);
+    }
+
+  g_assert_true (pango_cairo_font_map_get_resolution (PANGO_CAIRO_FONT_MAP (fontmap)) == 96.0);
+
+  return fontmap;
+}
+
